import unittest

from etsyv3.enums import ListingRequestState, WhenMade, WhoMade
from etsyv3.models import UpdateListingRequest
from etsyv3.models.listing_request import (
    CreateDraftListingRequest,
    UpdateListingInventoryRequest,
)
from etsyv3.models.product import Product
from etsyv3.util import todict


class TestListingRequest(unittest.TestCase):
    def test_nullable(self):
        listing = UpdateListingRequest(
            tags=[], title="A title!", who_made=WhoMade.I_DID
        )
        self.assertEqual(len(listing.get_nulled()), 1)
        self.assertEqual(3, len(listing.get_dict()))

    def test_required_missing(self):
        with self.assertRaises(ValueError):
            listing = CreateDraftListingRequest(
                quantity=1, title="Another title!", description="Description"
            )

    def test_required(self):
        listing = CreateDraftListingRequest(
            quantity=1,
            title="Another title!",
            description="Description",
            price=1.10,
            who_made=WhoMade.I_DID,
            when_made=WhenMade.MADE_TO_ORDER,
            taxonomy_id=1,
            tags=[],
        )
        self.assertEqual(8, len(listing.get_dict()))
        self.assertTrue("_nullable" not in listing.get_dict().keys())
        self.assertTrue("nullable" not in listing.get_dict().keys())
        self.assertEqual("i_did", listing.get_dict()["who_made"])

    def test_listing_inventory_request(self):
        properties = [
            {
                "property_id": 1,
                "value_ids": [1],
                "scale_id": 1,
                "property_name": "string",
                "values": ["string"],
            }
        ]
        offerings = [{"price": 0, "quantity": 0, "is_enabled": True}]
        product1 = Product("PRODUCT1", properties, offerings)
        product2 = Product("PRODUCT2", properties, offerings)
        product3 = Product("PRODUCT3", properties, offerings)
        product_list = [product1, product2, product3]
        listing_inventory_request = UpdateListingInventoryRequest(product_list, 0, 0, 0)
        self.assertEqual(3, len(listing_inventory_request.products))
        self.assertEqual(
            "PRODUCT1", listing_inventory_request.get_dict()["products"][0]["sku"]
        )

    def test_create_listing_inventory_request_from_response(self):
        response = {
            "products": [
                {
                    "product_id": 1,
                    "sku": "string",
                    "is_deleted": True,
                    "offerings": [
                        {
                            "offering_id": 1,
                            "quantity": 0,
                            "is_enabled": True,
                            "is_deleted": True,
                            "price": {
                                "amount": 0,
                                "divisor": 0,
                                "currency_code": "string",
                            },
                        }
                    ],
                    "property_values": [
                        {
                            "property_id": 1,
                            "property_name": "string",
                            "scale_id": 1,
                            "scale_name": "string",
                            "value_ids": [1],
                            "values": ["string"],
                        }
                    ],
                }
            ],
            "price_on_property": [0],
            "quantity_on_property": [0],
            "sku_on_property": [0],
            "listing": {
                "listing_id": 1,
                "user_id": 1,
                "shop_id": 1,
                "title": "string",
                "description": "string",
                "state": "active",
                "creation_timestamp": 946684800,
                "created_timestamp": 946684800,
                "ending_timestamp": 946684800,
                "original_creation_timestamp": 946684800,
                "last_modified_timestamp": 946684800,
                "updated_timestamp": 946684800,
                "state_timestamp": 946684800,
                "quantity": 0,
                "shop_section_id": 1,
                "featured_rank": 0,
                "url": "string",
                "num_favorers": 0,
                "non_taxable": True,
                "is_taxable": True,
                "is_customizable": True,
                "is_personalizable": True,
                "personalization_is_required": True,
                "personalization_char_count_max": 0,
                "personalization_instructions": "string",
                "listing_type": "physical",
                "tags": [],
                "materials": [],
                "shipping_profile_id": 1,
                "processing_min": 0,
                "processing_max": 0,
                "who_made": "i_did",
                "when_made": "made_to_order",
                "is_supply": True,
                "item_weight": 0,
                "item_weight_unit": "oz",
                "item_length": 0,
                "item_width": 0,
                "item_height": 0,
                "item_dimensions_unit": "in",
                "is_private": True,
                "style": [],
                "file_data": "string",
                "has_variations": True,
                "should_auto_renew": True,
                "language": "string",
                "price": {},
                "taxonomy_id": 0,
            },
        }



from datetime import datetime, timedelta, timezone
from enum import Enum
from typing import Callable, List

import requests

from etsyv3.models import ListingFile, ListingProperty, Request, UpdateListingRequest
from etsyv3.models.listing_request import (
    CreateDraftListingRequest,
    UpdateListingInventoryRequest,
    UpdateListingPropertyRequest,
    UpdateVariationImagesRequest,
    UploadListingImageRequest,
)

ETSY_API_BASEURL = "https://openapi.etsy.com/v3/application/"


class ExpiredToken(Exception):
    pass


class BadRequest(Exception):
    pass


class Unauthorised(Exception):
    pass


class NotFound(Exception):
    pass


class InternalError(Exception):
    pass


class Forbidden(Exception):
    pass


class Conflict(Exception):
    pass


class SortOn(Enum):
    CREATED = "created"
    PRICE = "price"
    UPDATED = "updated"
    SCORE = "score"


class SortOrder(Enum):
    ASC = "asc"
    ASCENDING = "ascending"
    DESC = "desc"
    DESCENDING = "descending"
    UP = "up"
    DOWN = "down"


class Includes(Enum):
    SHIPPING = "Shipping"
    IMAGES = "Images"
    SHOP = "Shop"
    USER = "User"
    TRANSLATIONS = "Translations"
    INVENTORY = "Inventory"


class ListingState(Enum):
    ACTIVE = "active"
    INACTIVE = "inactive"
    SOLD_OUT = "sold_out"
    DRAFT = "draft"
    EXPIRED = "expired"


class Method(Enum):
    GET = 1
    POST = 2
    PUT = 3
    DELETE = 4
    PATCH = 5


class EtsyAPI:
    def __init__(
        self,
        keystring: str,
        token: str,
        refresh_token: str,
        expiry: datetime,
        refresh_save: Callable[[str, str, datetime], None] = None,
    ):
        self.session = requests.Session()
        self.token = token
        self.user_id = token.split(".")[0]
        self.refresh_token = refresh_token
        self.keystring = keystring
        self.session.headers = {
            "Accept": "application/json",
            "x-api-key": keystring,
            "Authorization": "Bearer " + self.token,
        }
        self.expiry = expiry
        self.refresh_save = refresh_save

    @staticmethod
    def _generate_get_uri(uri, **kwargs):
        if kwargs == {} or kwargs is None:
            return uri
        params = "&".join(
            [f"{key}={value}" for key, value in kwargs.items() if value is not None]
        )
        uri = f"{uri}?{params}" if params != "" else uri
        return uri

    def _issue_request(
        self,
        uri,
        method: Method = Method.GET,
        request_payload: Request = None,
        **kwargs,
    ):
        if (
            method != Method.GET and method != Method.DELETE
        ) and request_payload is None:
            raise ValueError
        if (
            datetime.utcnow()
            < self.expiry
        ):
            if method == Method.GET:
                uri_full = EtsyAPI._generate_get_uri(uri, **kwargs)
                return_val = self.session.get(uri_full)
            elif method == method.PUT:
                return_val = self.session.put(uri, json=request_payload.get_dict())
            elif method == method.POST and isinstance(request_payload, UploadListingImageRequest):
                return_val = self.session.post(
                    uri,
                    files=request_payload.image,
                    data=request_payload.data
                )
            elif method == method.POST:
                return_val = self.session.post(uri, json=request_payload.get_dict())
            elif method == method.PATCH:
                return_val = self.session.patch(uri, json=request_payload.get_dict())
            else:
                return_val = self.session.delete(uri)
            if return_val.status_code == 400:
                raise BadRequest(return_val.json())
            elif return_val.status_code == 401:
                raise Unauthorised(return_val.json())
            elif return_val.status_code == 403:
                raise Forbidden(return_val.json())
            elif return_val.status_code == 409:
                raise Conflict(return_val.json())
            elif return_val.status_code == 404:
                raise NotFound(return_val.json())
            elif return_val.status_code == 500:
                raise InternalError(return_val.json())
            return return_val.json()

        else:
            self.refresh()
            return self._issue_request(uri, **kwargs)

    def get_buyer_taxonomy_nodes(self):
        uri = f"{ETSY_API_BASEURL}/buyer-taxonomy/nodes"
        return self._issue_request(uri)

    def get_properties_by_buyer_taxonomy_id(self, taxonomy_id):
        uri = f"{ETSY_API_BASEURL}/buyer-taxonomy/nodes/{taxonomy_id}/properties"
        return self._issue_request(uri)

    def get_seller_taxonomy_nodes(self):
        uri = f"{ETSY_API_BASEURL}/seller-taxonomy/nodes"
        return self._issue_request(uri)

    def get_properties_by_taxonomy_id(self, taxonomy_id):
        uri = f"{ETSY_API_BASEURL}/seller-taxonomy/nodes/{taxonomy_id}/properties"
        return self._issue_request(uri)

    def create_draft_listing(self, shop_id, listing: CreateDraftListingRequest):
        uri = f"{ETSY_API_BASEURL}/shops/{shop_id}/listings"
        return self._issue_request(uri, method=Method.POST, request_payload=listing)

    def get_listings_by_shop(
        self,
        shop_id,
        state: ListingState = None,
        limit: int = None,
        offset: int = None,
        sort_on: SortOn = None,
        sort_order: SortOrder = None,
        includes: List[Includes] = None,
    ):
        uri = f"{ETSY_API_BASEURL}/shops/{shop_id}/listings"
        kwargs = {
            "state": state.value if state is not None else None,
            "limit": limit,
            "offset": offset,
            "sort_on": sort_on.value if sort_on is not None else None,
            "sort_order": sort_order.value if sort_order is not None else None,
            "includes": [x.value for x in includes] if includes is not None else None,
        }
        return self._issue_request(uri, **kwargs)

    def delete_listing(self, listing_id: int):
        uri = f"{ETSY_API_BASEURL}/listings/{listing_id}"
        return self._issue_request(uri, method=Method.DELETE)

    def get_listing(self, listing_id: int, includes: List[Includes] = None):
        uri = f"{ETSY_API_BASEURL}/listings/{listing_id}"
        kwargs = {
            "includes": [x.value for x in includes] if includes is not None else None
        }
        return self._issue_request(uri, **kwargs)

    def find_all_listings_active(
        self,
        limit: int = None,
        offset: int = None,
        keywords: str = None,
        sort_on: SortOn = None,
        sort_order: SortOrder = None,
        min_price: float = None,
        max_price: float = None,
        shop_location: str = None,
    ):
        # not implementing taxonomy ids because I don't know what they are
        uri = f"{ETSY_API_BASEURL}/listings/active"
        kwargs = {
            "limit": limit,
            "offset": offset,
            "keywords": keywords,
            "sort_on": sort_on.value if sort_on is not None else None,
            "sort_order": sort_order.value if sort_order is not None else None,
            "min_price": min_price,
            "max_price": max_price,
            "shop_location": shop_location,
        }
        return self._issue_request(uri, **kwargs)

    def find_all_active_listings_by_shop(
        self,
        shop_id,
        limit: int = None,
        sort_on: SortOn = None,
        sort_order: SortOrder = None,
        offset: int = None,
        keywords: str = None,
    ):
        uri = f"{ETSY_API_BASEURL}/shops/{shop_id}/listings/active"
        kwargs = {
            "limit": limit,
            "sort_on": sort_on.value if sort_on is not None else None,
            "sort_order": sort_order.value if sort_order is not None else None,
            "offset": offset,
            "keywords": keywords,
        }
        return self._issue_request(uri, **kwargs)

    def get_listings_by_listing_ids(
        self, listing_ids: List[int], includes: List[Includes] = None
    ):
        uri = f"{ETSY_API_BASEURL}/listings/batch"
        kwargs = {
            "listing_ids": listing_ids,
            "includes": [x.value for x in includes] if includes is not None else None,
        }
        return self._issue_request(uri, **kwargs)

    def get_featured_listings_by_shop(self, shop_id, limit=None, offset=None):
        uri = f"{ETSY_API_BASEURL}/shops/{shop_id}/listings/featured"
        kwargs = {"limit": limit, "offset": offset}
        return self._issue_request(uri, **kwargs)

    def delete_listing_property(self, shop_id: int, listing_id: int, property_id: int):
        uri = f"{ETSY_API_BASEURL}/shops/{shop_id}/listings/{listing_id}/properties/{property_id}"
        return self._issue_request(uri, method=Method.DELETE)

    def update_listing_property(
        self,
        shop_id: int,
        listing_id: int,
        property_id: int,
        listing_property: UpdateListingPropertyRequest,
    ):
        uri = f"{ETSY_API_BASEURL}/shops/{shop_id}/listings/{listing_id}/properties/{property_id}"
        return self._issue_request(uri, method=Method.PUT, request_payload=listing_property)

    def get_listing_property(self, listing_id: int, property_id: int):
        # not in production yet
        uri = f"{ETSY_API_BASEURL}/listings/{listing_id}/properties/{property_id}"
        raise NotImplementedError

    def get_listing_properties(self, shop_id: int, listing_id: int):
        uri = f"{ETSY_API_BASEURL}/shops/{shop_id}/listings/{listing_id}/properties"
        return self._issue_request(uri)

    def update_listing(
        self, shop_id: int, listing_id: int, listing: UpdateListingRequest
    ):
        uri = f"{ETSY_API_BASEURL}/shops/{shop_id}/listings/{listing_id}"
        return self._issue_request(uri, method=Method.PATCH, request_payload=listing)

    def get_listings_by_shop_receipt(
        self, shop_id: int, receipt_id: int, limit=None, offset=None
    ):
        uri = f"{ETSY_API_BASEURL}/shops/{shop_id}/receipts/{receipt_id}/listings"
        kwargs = {"limit": limit, "offset": offset}
        return self._issue_request(uri, **kwargs)

    def get_listings_by_shop_section_id(
        self,
        shop_id: int,
        shop_section_ids: List[int],
        limit: int = None,
        offset: int = None,
        sort_on: SortOn = None,
        sort_order: SortOrder = None,
    ):
        uri = f"{ETSY_API_BASEURL}/shops/{shop_id}/shop-sections/listings"
        kwargs = {
            "shop_section_ids": shop_section_ids,
            "limit": limit,
            "offset": offset,
            "sort_on": sort_on.value if sort_on is not None else None,
            "sort_order": sort_order.value if sort_order is not None else None,
        }
        return self._issue_request(uri, **kwargs)

    def delete_listing_file(self, shop_id: int, listing_id: int, listing_file_id: int):
        uri = f"{ETSY_API_BASEURL}/shops/{shop_id}/listings/{listing_id}/files/{listing_file_id}"
        return self._issue_request(uri, method=Method.DELETE)

    def get_listing_file(self, shop_id: int, listing_id: int, listing_file_id: int):
        uri = f"{ETSY_API_BASEURL}/shops/{shop_id}/listings/{listing_id}/files/{listing_file_id}"
        return self._issue_request(uri)

    def get_all_listing_files(self, shop_id: int, listing_id: int):
        uri = f"{ETSY_API_BASEURL}/shops/{shop_id}/listings/{listing_id}/files"
        return self._issue_request(uri)

    def upload_listing_file(
        self, shop_id: int, listing_id: int, listing_file: ListingFile
    ):
        uri = f"{ETSY_API_BASEURL}/shops/{shop_id}/listings/{listing_id}/files"
        raise NotImplementedError

    def delete_listing_image(
        self, shop_id: int, listing_id: int, listing_image_id: int
    ):
        uri = f"{ETSY_API_BASEURL}/shops/{shop_id}/listings/{listing_id}/images/{listing_image_id}"
        return self._issue_request(uri, method=Method.DELETE)

    def get_listing_image(self, listing_id: int, listing_image_id: int):
        uri = f"{ETSY_API_BASEURL}/listings/{listing_id}/images/{listing_image_id}"
        return self._issue_request(uri)

    def get_listing_images(self, listing_id: int):
        uri = f"{ETSY_API_BASEURL}/listings/{listing_id}/images"
        return self._issue_request(uri)

    def upload_listing_image(
        self,
        shop_id: int,
        listing_id: int,
        listing_image: UploadListingImageRequest
    ):
        uri = f"{ETSY_API_BASEURL}/shops/{shop_id}/listings/{listing_id}/images"
        return self._issue_request(uri, method=Method.POST, request_payload=listing_image)

    def get_listing_inventory(self, listing_id):
        uri = f"{ETSY_API_BASEURL}/listings/{listing_id}/inventory"
        return self._issue_request(uri)

    def update_listing_inventory(
        self, listing_id: int, listing_inventory: UpdateListingInventoryRequest
    ):
        uri = f"{ETSY_API_BASEURL}/listings/{listing_id}/inventory"
        return self._issue_request(uri, Method.PUT, listing_inventory)

    def get_listing_offering(
        self, listing_id: int, product_id: int, product_offering_id: int
    ):
        uri = f"{ETSY_API_BASEURL}/listings/{listing_id}/products/{product_id}/offerings/{product_offering_id}"
        return self._issue_request(uri)

    def get_listing_product(self, listing_id: int, product_id: int):
        uri = (
            f"{ETSY_API_BASEURL}/listings/{listing_id}/inventory/products/{product_id}"
        )
        return self._issue_request(uri)

    def create_listing_translation(self):
        raise NotImplementedError

    def get_listing_translation(self, shop_id: int, listing_id: int, language: str):
        uri = f"{ETSY_API_BASEURL}/shops/{shop_id}/listings/{listing_id}/translations/{language}"
        return self._issue_request(uri)

    def update_listing_translation(self):
        raise NotImplementedError

    def get_listing_variation_images(self, shop_id: int, listing_id: int):
        uri = (
            f"{ETSY_API_BASEURL}/shops/{shop_id}/listings/{listing_id}/variation-images"
        )
        return self._issue_request(uri)

    def update_variation_images(
        self,
        shop_id: int,
        listing_id: int,
        variation_images: UpdateVariationImagesRequest,
    ):
        uri = (
            f"{ETSY_API_BASEURL}/shops/{shop_id}/listings/{listing_id}/variation-images"
        )
        return self._issue_request(
            uri, method=Method.POST, request_payload=variation_images
        )

    def ping(self):
        uri = f"{ETSY_API_BASEURL}/openapi-ping"
        return self._issue_request(uri)

    def token_scopes(self):
        uri = f"{ETSY_API_BASEURL}/scopes"
        return self._issue_request(uri)

    def get_shop_payment_account_ledger_entry(self, shop_id: int, ledger_entry_id: int):
        uri = f"{ETSY_API_BASEURL}/shops/{shop_id}/payment-account/ledger-entries/{ledger_entry_id}"
        return self._issue_request(uri)

    def get_shop_payment_account_ledger_entries(
        self,
        shop_id: int,
        min_created: int,
        max_created: int,
        limit: int = None,
        offset: int = None,
    ):
        uri = f"{ETSY_API_BASEURL}/shops/{shop_id}/payment-account/ledger-entries"
        kwargs = {
            "min_created": min_created,
            "max_created": max_created,
            "limit": limit,
            "offset": offset,
        }
        return self._issue_request(uri, **kwargs)

    def get_payment_account_ledger_entry_payments(
        self, shop_id: int, ledger_entry_ids: List[int]
    ):
        uri = f"{ETSY_API_BASEURL}/shops/{shop_id}/payment-account/ledger-entries/payments"
        kwargs = {"ledger_entry_ids": ledger_entry_ids}
        return self._issue_request(uri, **kwargs)

    def get_shop_payment_by_receipt_id(self, shop_id: int, receipt_id: int):
        uri = f"{ETSY_API_BASEURL}/shops/{shop_id}/receipts/{receipt_id}/payments"
        return self._issue_request(uri)

    def get_payments(self, shop_id: int, payment_ids: List[int]):
        uri = f"{ETSY_API_BASEURL}/shops/{shop_id}/payments"
        kwargs = {"payment_ids": payment_ids}
        return self._issue_request(uri, **kwargs)

    def get_shop_receipt(self, shop_id: int, receipt_id: int):
        uri = f"{ETSY_API_BASEURL}/shops/{shop_id}/receipts/{receipt_id}"
        return self._issue_request(uri)

    def update_shop_receipt(self):
        raise NotImplementedError

    def get_shop_receipts(self, shop_id: int, limit: int = None, offset: int = None, was_paid: bool = True,
                          was_shipped: bool = False, was_canceled: bool = None):
        uri = f"{ETSY_API_BASEURL}/shops/{shop_id}/receipts"
        kwargs = {"limit": limit, "offset": offset, "was_paid": was_paid, "was_shipped": was_shipped,
                  "was_canceled": was_canceled}
        return self._issue_request(uri, **kwargs)

    def create_receipt_shipment(self):
        raise NotImplementedError

    def get_shop_receipt_transactions_by_listing(
        self, shop_id: int, listing_id: int, limit: int = None, offset: int = None
    ):
        uri = f"{ETSY_API_BASEURL}/shops/{shop_id}/listings/{listing_id}/transactions"
        kwargs = {"limit": limit, "offset": offset}
        return self._issue_request(uri, **kwargs)

    def get_shop_receipt_transactions_by_receipt(self, shop_id: int, receipt_id: int):
        uri = f"{ETSY_API_BASEURL}/shops/{shop_id}/receipts/{receipt_id}/transactions"
        return self._issue_request(uri)

    def get_shop_receipt_transaction(self, shop_id: int, transaction_id: int):
        uri = f"{ETSY_API_BASEURL}/shops/{shop_id}/transactions/{transaction_id}"
        return self._issue_request(uri)

    def get_shop_receipt_transactions_by_shop(
        self, shop_id: int, limit: int = None, offset: int = None
    ):
        uri = f"{ETSY_API_BASEURL}/shops/{shop_id}/transactions"
        kwargs = {"limit": limit, "offset": offset}
        return self._issue_request(uri, **kwargs)

    def get_reviews_by_listing(
        self, listing_id: int, limit: int = None, offset: int = None
    ):
        uri = f"{ETSY_API_BASEURL}/listings/{listing_id}/reviews"
        kwargs = {"limit": limit, "offset": offset}
        return self._issue_request(uri, **kwargs)

    def get_reviews_by_shop(self, shop_id: int, limit: int = None, offset: int = None):
        uri = f"{ETSY_API_BASEURL}/shops/{shop_id}/reviews"
        kwargs = {"limit": limit, "offset": offset}
        return self._issue_request(uri, **kwargs)

    def get_shipping_carriers(self, origin_country_iso: str):
        uri = f"{ETSY_API_BASEURL}/shipping-carriers"
        kwargs = {"origin_country_iso": origin_country_iso}
        return self._issue_request(uri, **kwargs)

    def create_shop_shipping_profile(self):
        raise NotImplementedError

    def get_shop_shipping_profiles(self, shop_id: int):
        uri = f"{ETSY_API_BASEURL}/shops/{shop_id}/shipping-profiles"
        return self._issue_request(uri)

    def delete_shop_shipping_profile(self, shop_id: int, shipping_profile_id: int):
        uri = f"{ETSY_API_BASEURL}/shops/{shop_id}/shipping-profiles/{shipping_profile_id}"
        return self._issue_request(uri, method=Method.DELETE)

    def get_shop_shipping_profile(self, shop_id: int, shipping_profile_id: int):
        uri = f"{ETSY_API_BASEURL}/shops/{shop_id}/shipping-profiles/{shipping_profile_id}"
        return self._issue_request(uri)

    def update_shop_shipping_profile(self):
        raise NotImplementedError

    def create_shop_shipping_profile_destination(self):
        raise NotImplementedError

    def get_shop_shipping_profile_destinations_by_shipping_profile(
        self,
        shop_id: int,
        shipping_profile_id: int,
        limit: int = None,
        offset: int = None,
    ):
        uri = f"{ETSY_API_BASEURL}/shops/{shop_id}/shipping-profiles/{shipping_profile_id}/destinations"
        kwargs = {"limit": limit, "offset": offset}
        return self._issue_request(uri, **kwargs)

    def delete_shop_shipping_profile_destination(
        self,
        shop_id: int,
        shipping_profile_id: int,
        shipping_profile_destination_id: int,
    ):
        uri = f"{ETSY_API_BASEURL}/shops/{shop_id}/shipping-profiles/{shipping_profile_id}/destinations/{shipping_profile_destination_id}"
        return self._issue_request(uri, method=Method.DELETE)

    def update_shop_shipping_profile_destination(self):
        raise NotImplementedError

    def create_shop_shipping_profile_upgrade(self):
        raise NotImplementedError

    def get_shop_shipping_profile_upgrades(
        self, shop_id: int, shipping_profile_id: int
    ):
        uri = f"{ETSY_API_BASEURL}/shops/{shop_id}/shipping-profiles/{shipping_profile_id}/upgrades"
        return self._issue_request(uri)

    def delete_shop_shipping_profile_upgrade(
        self, shop_id: int, shipping_profile_id: int, upgrade_id: int
    ):
        uri = f"{ETSY_API_BASEURL}/shops/{shop_id}/shipping-profiles/{shipping_profile_id}/upgrades/{upgrade_id}"
        return self._issue_request(uri, method=Method.DELETE)

    def update_shop_shipping_profile_upgrade(self):
        raise NotImplementedError

    def get_shop(self, shop_id: int):
        uri = f"{ETSY_API_BASEURL}/shops/{shop_id}"
        return self._issue_request(uri)

    def update_shop(self):
        raise NotImplementedError

    def get_shop_by_owner_user_id(self, user_id: int):
        uri = f"{ETSY_API_BASEURL}/users/{user_id}/shops"
        return self._issue_request(uri)

    def find_shops(self, shop_name: str, limit: int = None, offset: int = None):
        uri = f"{ETSY_API_BASEURL}/shops"
        kwargs = {"shop_name": shop_name, "limit": limit, "offset": offset}
        return self._issue_request(uri, **kwargs)

    def get_shop_production_partners(self, shop_id: int):
        uri = f"{ETSY_API_BASEURL}/shops/{shop_id}/production-partners"
        return self._issue_request(uri)

    def create_shop_section(self):
        raise NotImplementedError

    def get_shop_sections(self, shop_id: int):
        uri = f"{ETSY_API_BASEURL}/shops/{shop_id}/sections"
        return self._issue_request(uri)

    def delete_shop_section(self, shop_id: int, shop_section_id: int):
        uri = f"{ETSY_API_BASEURL}/shops/{shop_id}/sections/{shop_section_id}"
        return self._issue_request(uri, method=Method.DELETE)

    def get_shop_section(self, shop_id: int, shop_section_id: int):
        uri = f"{ETSY_API_BASEURL}/shops/{shop_id}/sections/{shop_section_id}"
        return self._issue_request(uri)

    def update_shop_section(self):
        raise NotImplementedError

    def get_user(self, user_id):
        uri = f"{ETSY_API_BASEURL}/users/{user_id}"
        return self._issue_request(uri)

    def get_authenticated_user(self):
        uri = f"{ETSY_API_BASEURL}/users/{self.user_id}"
        return self._issue_request(uri)

    def delete_user_address(self):
        raise NotImplementedError

    def get_user_address(self):
        raise NotImplementedError

    def get_user_addresses(self, limit: int = None, offset: int = None):
        uri = f"{ETSY_API_BASEURL}/user/addresses"
        kwargs = {"limit": limit, "offset": offset}
        return self._issue_request(uri, **kwargs)

    def refresh(self):
        data = {
            "grant_type": "refresh_token",
            "client_id": self.keystring,
            "refresh_token": self.refresh_token,
        }
        del self.session.headers["Authorization"]
        r = self.session.post("https://api.etsy.com/v3/public/oauth/token", json=data)
        refreshed = r.json()
        self.token = refreshed["access_token"]
        self.refresh_token = refreshed["refresh_token"]
        tmp_expiry = datetime.utcnow() + timedelta(seconds=refreshed["expires_in"])
        self.expiry = tmp_expiry
        self.session.headers["Authorization"] = "Bearer " + self.token
        if self.refresh_save is not None:
            self.refresh_save(self.token, self.refresh_token, self.expiry)
        return self.token, self.refresh_token, self.expiry



import unittest

from etsyv3.util.auth import AuthHelper


class TestAuthHelper(unittest.TestCase):
    def test__generate_challenge(self):
        code_verifier = "vvkdljkejllufrvbhgeiegrnvufrhvrffnkvcknjvfid"
        code_challenge = AuthHelper._generate_challenge(code_verifier)
        code_challenge_expected = "DSWlW2Abh-cf8CeLL8-g3hQ2WQyYdKyiu83u_s7nRhI"
        self.assertEqual(code_challenge, code_challenge_expected)



import base64
import hashlib
import secrets

from requests_oauthlib import OAuth2Session


class AuthHelper:
    def __init__(
        self, keystring, redirect_uri, scopes=None, code_verifier=None, state=None
    ):
        self.keystring = keystring
        self.redirect_url = redirect_uri
        self.scopes = scopes
        if code_verifier is None:
            self.code_verifier = secrets.token_urlsafe(32)
        else:
            self.code_verifier = code_verifier
        self.code_challenge = AuthHelper._generate_challenge(self.code_verifier)
        self.oauth = OAuth2Session(
            keystring, redirect_uri=self.redirect_url, scope=scopes
        )
        self.state = secrets.token_urlsafe(16) if state is None else state
        self.auth_code = None
        self.token = None

    def get_auth_code(self):
        authorisation_url, state = self.oauth.authorization_url(
            "https://www.etsy.com/oauth/connect",
            state=self.state,
            code_challenge=self.code_challenge,
            code_challenge_method="S256",
        )
        return authorisation_url, state

    def set_authorisation_code(self, code, state):
        if state == self.state:
            self.auth_code = code
        else:
            raise

    def get_access_token(self):
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/x-www-form-urlencoded",
            "x-api-key": self.keystring,
        }
        self.token = self.oauth.fetch_token(
            "https://api.etsy.com/v3/public/oauth/token",
            code=self.auth_code,
            code_verifier=self.code_verifier,
            include_client_id=True,
            headers=headers,
        )
        return self.token

    @staticmethod
    def _generate_challenge(code_verifier):
        m = hashlib.sha256(code_verifier.encode("utf-8"))
        b64_encode = base64.urlsafe_b64encode(m.digest()).decode("utf-8")
        # per https://docs.python.org/3/library/base64.html, there may be a trailing '=' - get rid of it
        return b64_encode.split("=")[0]



from .listing_file import ListingFile
from .listing_inventory import ListingInventory
from .listing_property import ListingProperty
from .listing_request import Request, UpdateListingRequest, UpdateVariationImagesRequest



from enum import Enum


def todict(obj, classkey=None, nullable=None):
    if nullable is None:
        nullable = []
    if isinstance(obj, dict):
        data = {}
        for (k, v) in obj.items():
            data[k] = todict(v, classkey)
        return data
    elif Enum in obj.__class__.__mro__:
        return todict(obj.value)
    elif hasattr(obj, "_ast"):
        return todict(obj._ast())
    elif hasattr(obj, "__iter__") and not isinstance(obj, str):
        return [todict(v, classkey) for v in obj]
    elif hasattr(obj, "__dict__"):
        data = dict(
            [
                (key, todict(value, classkey))
                if key not in nullable and value not in [[], "" or 0]
                else (key, None)
                for key, value in obj.__dict__.items()
                if not callable(value) and not key.startswith("_") and value is not None
            ]
        )
        if classkey is not None and hasattr(obj, "__class__"):
            data[classkey] = obj.__class__.__name__
        return data
    else:
        return obj



from enum import Enum
from typing import Any, Dict, List

from etsyv3.enums import (
    ItemDimensionsUnit,
    ItemWeightUnit,
    ListingRequestState,
    ListingType,
    WhenMade,
    WhoMade,
)
from etsyv3.models.product import Product
from etsyv3.util import todict


class Request:
    def __init__(self, nullable: List[str] = None, mandatory: List[str] = None):
        self._nullable = nullable
        self._mandatory = mandatory
        if not self.check_mandatory():
            raise ValueError

    def check_mandatory(self) -> bool:
        if self.mandatory is not None:
            for key in self.mandatory:
                try:
                    if self.__dict__[key] is None:
                        return False
                except Exception:
                    return False
        return True

    def get_nulled(self) -> List[str]:
        return [
            key
            for key, value in self.__dict__.items()
            if key in self._nullable and (value == [] or value == "" or value == 0)
        ]

    def get_dict(self) -> Dict[str, Any]:
        nulled = self.get_nulled()
        return todict(self, nullable=nulled)


class CreateDraftListingRequest(Request):
    nullable = [
        "shipping_profile_id",
        "materials",
        "shop_section_id",
        "processing_min",
        "processing_max",
        "tags",
        "styles",
        "item_weight",
        "item_length",
        "item_width",
        "item_height",
        "item_weight_unit",
        "item_dimensions_unit",
        "production_partner_ids",
        "image_ids",
    ]
    mandatory = [
        "quantity",
        "title",
        "description",
        "price",
        "who_made",
        "when_made",
        "taxonomy_id",
    ]

    def __init__(
        self,
        quantity: int = None,
        title: str = None,
        description: str = None,
        price: float = None,
        who_made: WhoMade = None,
        when_made: WhenMade = None,
        taxonomy_id: int = None,
        shipping_profile_id: int = None,
        materials: List[str] = None,
        shop_section_id: int = None,
        processing_min: int = None,
        processing_max: int = None,
        tags: List[str] = None,
        styles: List[str] = None,
        item_weight: float = None,
        item_length: float = None,
        item_width: float = None,
        item_height: float = None,
        item_weight_unit: ItemWeightUnit = None,
        item_dimensions_unit: ItemDimensionsUnit = None,
        is_personalizable: bool = None,
        personalization_is_required: bool = None,
        personalization_char_count_max: int = None,
        personalization_instructions: str = None,
        production_partner_ids: int = None,
        image_ids: List[int] = None,
        is_supply: bool = None,
        is_customizable: bool = None,
        should_auto_renew: bool = None,
        is_taxable: bool = None,
        listing_type: ListingType = None,
    ):
        self.quantity = quantity
        self.title = title
        self.description = description
        self.price = price
        self.who_made = who_made
        self.when_made = when_made
        self.taxonomy_id = taxonomy_id
        self.shipping_profile_id = shipping_profile_id
        self.materials = materials
        self.shop_section_id = shop_section_id
        self.processing_min = processing_min
        self.processing_max = processing_max
        self.tags = tags
        self.styles = styles
        self.item_weight = item_weight
        self.item_length = item_length
        self.item_width = item_width
        self.item_height = item_height
        self.item_weight_unit = item_weight_unit
        self.item_dimensions_unit = item_dimensions_unit
        self.is_personalizable = is_personalizable
        self.personalization_is_required = personalization_is_required
        self.personalization_char_count_max = personalization_char_count_max
        self.personalization_instructions = personalization_instructions
        self.production_partner_ids = production_partner_ids
        self.image_ids = image_ids
        self.is_supply = is_supply
        self.is_customizable = is_customizable
        self.should_auto_renew = should_auto_renew
        self.is_taxable = is_taxable
        self.listing_type = listing_type
        super().__init__(
            nullable=CreateDraftListingRequest.nullable,
            mandatory=CreateDraftListingRequest.mandatory,
        )


class UpdateListingRequest(Request):
    nullable = [
        "materials",
        "shipping_profile_id",
        "shop_section_id",
        "item_weight",
        "item_length",
        "item_width",
        "item_height",
        "item_weight_unit",
        "item_dimensions_unit",
        "tags",
        "featured_rank",
        "production_partner_ids",
        "type",
    ]

    mandatory = []

    def __init__(
        self,
        image_ids: List[str] = None,
        title: str = None,
        description: str = None,
        materials: List[str] = None,
        should_auto_renew: bool = None,
        shipping_profile_id: int = None,
        shop_section_id: int = None,
        item_weight: float = None,
        item_length: float = None,
        item_width: float = None,
        item_height: float = None,
        item_weight_unit: ItemWeightUnit = None,
        item_dimensions_unit: ItemDimensionsUnit = None,
        is_taxable: bool = None,
        taxonomy_id: int = None,
        tags: List[str] = None,
        who_made: WhoMade = None,
        when_made: WhenMade = None,
        featured_rank: int = None,
        is_personalizable: bool = None,
        personalization_is_required: bool = None,
        personalization_char_count_max: int = None,
        personalization_instructions: str = None,
        state: ListingRequestState = None,
        is_supply: bool = None,
        production_partner_ids: List[int] = None,
        listing_type: ListingType = None,
    ):
        self.image_ids = image_ids
        self.title = title
        self.description = description
        self.materials = materials
        self.should_auto_renew = should_auto_renew
        self.shipping_profile_id = shipping_profile_id
        self.shop_section_id = shop_section_id
        self.item_weight = item_weight
        self.item_length = item_length
        self.item_width = item_width
        self.item_height = item_height
        self.item_weight_unit = item_weight_unit
        self.item_dimensions_unit = item_dimensions_unit
        self.is_taxable = is_taxable
        self.taxonomy_id = taxonomy_id
        self.tags = tags
        self.who_made = who_made
        self.when_made = when_made
        self.featured_rank = featured_rank
        self.is_personalizable = is_personalizable
        self.personalization_is_required = personalization_is_required
        self.personalization_char_count_max = personalization_char_count_max
        self.personalization_instructions = personalization_instructions
        self.state = state
        self.is_supply = is_supply
        self.production_partner_ids = production_partner_ids
        self.listing_type = listing_type
        super().__init__(nullable=UpdateListingRequest.nullable)


class UpdateListingInventoryRequest(Request):
    nullable = []
    mandatory = ["products"]

    def __init__(
        self,
        products: List[Product],
        price_on_property=None,
        quantity_on_property=None,
        sku_on_property=None,
    ):
        self.products = products
        self.price_on_property = price_on_property
        self.quantity_on_property = quantity_on_property
        self.sku_on_property = sku_on_property
        super().__init__(
            nullable=UpdateListingInventoryRequest.nullable,
            mandatory=UpdateListingInventoryRequest.mandatory,
        )

    @staticmethod
    def generate_request_from_inventory_response(response):
        products = []
        for product in response["products"]:
            property_values = product["property_values"]
            for prop_val in property_values:
                prop_val.pop("scale_name", None)
                prop_val.pop("value_pairs", None)
            offerings = product["offerings"]
            for offering in offerings:
                offering.pop("is_deleted", None)
                offering.pop("offering_id", None)
                offering["price"] = (
                    offering["price"]["amount"] / offering["price"]["divisor"]
                )
            products.append(Product(product["sku"], property_values, offerings))
        price_on_property = response["price_on_property"]
        quantity_on_property = response["quantity_on_property"]
        sku_on_property = response["sku_on_property"]
        return UpdateListingInventoryRequest(
            products, price_on_property, quantity_on_property, sku_on_property
        )


class UpdateVariationImagesRequest(Request):
    nullable = []
    mandatory = []

    def __init__(self, variation_images: List[Dict[str, Any]]):
        self.variation_images = variation_images
        super().__init__(
            nullable=UpdateVariationImagesRequest.nullable,
            mandatory=UpdateVariationImagesRequest.mandatory,
        )

    @staticmethod
    def generate_request_from_variation_images_response(response):
        variation_images = response["results"]
        for variation_image in variation_images:
            variation_image.pop("value", None)
        return UpdateVariationImagesRequest(variation_images)

class UpdateListingPropertyRequest(Request):
    nullable = []

    mandatory = [
        "value_ids",
        "values",
    ]

    def __init__(
        self,
        value_ids: List[int] = None,
        values: List[str] = None,
        scale_id: int = None,
    ):
        self.value_ids = value_ids
        self.values = values
        self.scale_id = scale_id
        super().__init__(
            nullable=UpdateListingPropertyRequest.nullable,
            mandatory=UpdateListingPropertyRequest.mandatory
        )


class UploadListingImageRequest(Request):
    nullable = ["image"]
    mandatory = []

    def __init__(
        self,
        image_bytes: bytes,
        listing_image_id=None,
        rank=None,
        overwrite=None,
        is_watermarked=None,
        alt_text=None,
    ):
        self.image = {
            "image": image_bytes
        }
        self.data = {
            "listing_image_id": listing_image_id,
            "rank": rank,
            "overwrite": overwrite,
            "is_watermarked": is_watermarked,
            "alt_text": alt_text,
        }
        
        super().__init__(
            nullable=UploadListingImageRequest.nullable,
            mandatory=UploadListingImageRequest.mandatory,
        )
    
    @staticmethod
    def generate_image_bytes_from_file(file):
        with open(file, "rb") as image:
            image_bytes = image.read()
        return image_bytes



from typing import Any, Dict, List


class Product:
    def __init__(
        self,
        sku: str,
        property_values: List[Dict[str, Any]],
        offerings: List[Dict[str, Any]],
    ):
        self.sku = sku
        self.property_values = property_values
        self.offerings = offerings



import unittest
from datetime import datetime, timedelta
from unittest import mock

import tests.mock_helpers
from etsyv3 import EtsyAPI
from etsyv3.etsy_api import ETSY_API_BASEURL, Unauthorised

EXPIRY_FUTURE = datetime.utcnow() + timedelta(hours=1)
EXPIRY_PAST = datetime.utcnow() - timedelta(hours=1)
KEYSTRING = ""
TOKEN = ""
REFRESH_TOKEN = ""


def fn_save(one, two, three):
    pass


class TestEtsyAPI(unittest.TestCase):
    def test__generate_get_uri_one_value(self):
        test_dict = {"limit": None, "offset": 100}
        url = EtsyAPI._generate_get_uri(ETSY_API_BASEURL, **test_dict)
        expect_url = f"{ETSY_API_BASEURL}?offset=100"
        self.assertEqual(expect_url, url)

    def test__generate_get_uri_two_values(self):
        test_dict = {"limit": 100, "offset": 100}
        url = EtsyAPI._generate_get_uri(ETSY_API_BASEURL, **test_dict)
        expect_url = f"{ETSY_API_BASEURL}?limit=100&offset=100"
        self.assertEqual(expect_url, url)

    def test__generate_get_uri_none_values_in_kwargs(self):
        test_dict = {"limit": None, "offset": None}
        url = EtsyAPI._generate_get_uri(ETSY_API_BASEURL, **test_dict)
        expect_url = f"{ETSY_API_BASEURL}"
        self.assertEqual(expect_url, url)

    def test__generate_get_uri_no_kwargs(self):
        url = EtsyAPI._generate_get_uri(ETSY_API_BASEURL)
        expect_url = f"{ETSY_API_BASEURL}"
        self.assertEqual(expect_url, url)

    @mock.patch(
        "requests.Session.get", side_effect=tests.mock_helpers.mocked_requests_get
    )
    def test_get_user_unauthorised(self, mock_get):
        etsy = EtsyAPI(KEYSTRING, TOKEN, REFRESH_TOKEN, EXPIRY_FUTURE, fn_save)
        with self.assertRaises(Unauthorised) as context:
            etsy.get_user("UNAUTHORIZED")

    @mock.patch(
        "requests.Session.get", side_effect=tests.mock_helpers.mocked_requests_get
    )
    def test_get_self(self, mock_get):
        etsy = EtsyAPI(KEYSTRING, TOKEN, REFRESH_TOKEN, EXPIRY_FUTURE, fn_save)
        etsy.get_authenticated_user()
        mock_get.assert_called()



from enum import Enum


class ItemWeightUnit(Enum):
    OZ = "oz"
    LB = "lb"
    G = "g"
    KG = "kg"


class ItemDimensionsUnit(Enum):
    IN = "in"
    FT = "ft"
    MM = "mm"
    YD = "yd"
    INCHES = "inches"


class WhoMade(Enum):
    I_DID = "i_did"
    SOMEONE_ELSE = "someone_else"
    COLLECTIVE = "collective"


class WhenMade(Enum):
    MADE_TO_ORDER = "made_to_order"
    TWENTY_TWENTIES = "2020_2022"
    TWENTY_TENS = "2010_2019"
    TWENTY_OH_THREE_TO_NINE = "2003_2009"
    BEFORE_2003 = "before_2003"
    TWO_THOUSAND_TO_TWO = "2000_2002"
    NINETEEN_NINETIES = "1990s"
    NINETEEN_EIGHTIES = "1980s"
    NINETEEN_SEVENTIES = "1970s"
    NINETEEN_SIXTIES = "1960s"
    NINETEEN_FIFTIES = "1950s"
    NINETEEN_FORTIES = "1940s"
    NINETEEN_THIRTIES = "1930s"
    NINETEEN_TWENTIES = "1920s"
    NINETEEN_TENS = "1910s"
    NINETEEN_HUNDREDS = "1900s"
    EIGHTEEN_HUNDREDS = "1800s"
    SEVENTEEN_HUNDREDS = "1700s"
    BEFORE_1700 = "before_1700"


class ListingRequestState(Enum):
    ACTIVE = "active"
    INACTIVE = "inactive"


class ListingType(Enum):
    PHYSICAL = "physical"
    DOWNLOAD = "download"
    BOTH = "both"
